# coding=UTF-8
from __future__ import print_function, absolute_import, division

__all__ = [
    'FirankaError',
    'NotInDomainError',
]


class FirankaError(Exception):
    """
    Base class for firanka's exceptions
    """


class NotInDomainError(FirankaError, ValueError):
    """
    Requested index is beyond this domain
    """



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import math

from .base import Series
from ..ranges import REAL_SET


class ModuloSeries(Series):
    def __init__(self, series, *args, **kwargs):
        """
        Construct a modulo series
        :param series: base series to use
        :raise ValueError: invalid domain length
        """
        super(ModuloSeries, self).__init__(REAL_SET, *args, **kwargs)

        self.series = series
        self.period = self.series.domain.length()

        if self.period == 0:
            raise ValueError('Modulo series cannot have a period of 0')
        elif math.isinf(self.period):
            raise ValueError('Modulo series cannot have an infinite period')

    def _get_for(self, item):
        if item < 0:
            item = -(item // self.period) * self.period + item
        elif item > self.period:
            item = item - (item // self.period) * self.period
        elif item == self.period:
            item = 0

        return self.series._get_for(self.series.domain.start + item)



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import functools
import math

import six

__all__ = [
    'Range',
    'REAL_SET',
    'EMPTY_SET'
]


def _pre_range(fun):
    @functools.wraps(fun)
    def inner(self, arg, *args, **kwargs):
        if not isinstance(arg, Range):
            arg = Range(arg)
        return fun(self, arg, *args, **kwargs)

    return inner


class Range(object):
    """
    Range of real numbers. Immutable.
    """

    def translate(self, x):
        return Range(self.start + x, self.stop + x, self.left_inc,
                     self.right_inc)

    def __init__(self, *args):
        """
        Create like:

        * Range('<a;b>')
        * Range(a, b, is_left_closed_, is_right_closed)
        * Range(a, b) - will have both sides closed, unless one is inf
        * Range(slice(a, b)) - will have both sides closed, unless one is None

        :param args:
        """
        if len(args) == 1:
            rs, = args
            if isinstance(rs, type(self)):
                args = rs.start, rs.stop, rs.left_inc, rs.right_inc
            elif isinstance(rs, slice):
                start = rs.start if rs.start is not None else float('-inf')
                stop = rs.stop if rs.stop is not None else float('+inf')
                args = start, stop, not math.isinf(start), not math.isinf(stop)
            else:
                if rs[0] not in '<(': raise ValueError(
                    'Must start with ( or <')
                if rs[-1] not in '>)': raise ValueError('Must end with ) or >')
                if ';' not in rs: raise ValueError('Separator ; required')

                start, stop = rs[1:-1].split(';')
                args = float(start), float(stop), rs[0] == '<', rs[-1] == '>'

        elif len(args) == 2:
            args = args[0], args[1], not math.isinf(args[0]), not math.isinf(
                args[1])

        q = lambda a, b, args: args[a] and math.isinf(args[b])

        if q(2, 0, args) or q(3, 1, args):
            raise ValueError('Set with sharp closing but infinity set')

        self.start, self.stop, self.left_inc, self.right_inc = args

    def __contains__(self, x):
        """
        :type x: index or a Range
        """
        if isinstance(x, six.string_types):
            x = Range(x)

        if isinstance(x, Range):
            if ((x.start == self.start) and (x.left_inc ^ self.left_inc)) \
                    or ((x.stop == self.stop) and (
                                x.right_inc ^ self.right_inc)):
                return False

            return (x.start >= self.start) and (x.stop <= self.stop)
        else:
            if x == self.start:
                return self.left_inc

            if x == self.stop:
                return self.right_inc

            return self.start < x < self.stop

    def is_empty(self):
        return (self.start == self.stop) and not (
            self.left_inc or self.right_inc)

    def length(self):
        return self.stop - self.start

    def __repr__(self):
        return 'Range(%s, %s, %s, %s)' % (
            repr(self.start), repr(self.stop), repr(self.left_inc),
            repr(self.right_inc))

    def __getitem__(self, item):
        if not isinstance(item, slice):
            raise ValueError('must be a slice')

        return self.intersection(Range(item))

    def __str__(self):
        return '%s%s;%s%s' % (
            '<' if self.left_inc else '(',
            self.start,
            self.stop,
            '>' if self.right_inc else ')',
        )

    @_pre_range
    def intersection(self, y):
        if self.start > y.start:
            return y.intersection(self)

        assert self.start <= y.start

        if (self.stop < y.start) or (y.stop < y.start):
            return EMPTY_SET

        if self.stop == y.start and not (self.right_inc and y.left_inc):
            return EMPTY_SET

        if self.start == y.start:
            start = self.start
            left_inc = self.left_inc and y.left_inc
        else:
            start = y.start
            left_inc = y.left_inc

        if self.stop == y.stop:
            stop = self.stop
            right_inc = self.right_inc and y.right_inc
        else:
            p, q = (self, y) if self.stop < y.stop else (y, self)
            stop = p.stop
            right_inc = p.right_inc and (stop in q)

        return Range(start, stop, left_inc, right_inc)

    @_pre_range
    def __eq__(self, other):
        if self.is_empty() and other.is_empty():
            return True
        return self.start == other.start and self.stop == other.stop and self.left_inc == other.left_inc and self.right_inc == other.right_inc

    def __hash__(self):
        return hash(self.start) ^ hash(self.stop)


EMPTY_SET = Range(0, 0, False, False)
REAL_SET = Range(float('-inf'), float('+inf'), False, False)



# coding=UTF-8
from __future__ import print_function, absolute_import, division

from .ranges import Range
from .series import Series


class BijectionMapping(object):
    def __init__(self, user2float, float2user):
        """
        :param user2float: callable/1 => float, mapping from your time system to floats
        :param float2user: callable/float => 1, mapping from float to your time systen
        """
        self.user2float = user2float
        self.float2user = float2user

    def to_float(self, user):
        return self.user2float(user)

    def to_user(self, flt):
        return self.float2user(flt)


class TimeProvidedSeries(Series):
    """
    If your time is something else than simple floats, this will help you out
    """

    def __init__(self, series, mapping, *args, **kwargs):
        """
        :param series: series to overlay
        """
        super(TimeProvidedSeries, self).__init__(series.domain, *args,
                                                 **kwargs)
        self.mapping = mapping
        self.series = series

    def _withmap(self, series):
        return TimeProvidedSeries(series, self.mapping)

    def __getitem__(self, item):
        if isinstance(item, slice):
            item = slice(
                float('-inf') if slice.start is None else self.to_float(
                    slice.start),
                float('+inf') if slice.stop is None else self.to_float(
                    slice.stop), )

            return self._withmap(self.series[item])
        elif isinstance(item, Range):
            return self._withmap(self.series[item])
        else:
            return self.series[self.mapping.to_float(item)]

    def _get_for(self, item):
        return self.series[self.l2f(item)]

    def join(self, series, fun):
        return TimeProvidedSeries(self.series.join(series, fun), self.mapping)



# coding=UTF-8
from __future__ import absolute_import

from .base import FunctionSeries, DiscreteSeries, Series
from .interpolations import LinearInterpolationSeries, \
    SCALAR_LINEAR_INTERPOLATOR
from .modulo import ModuloSeries

__all__ = [
    'FunctionSeries',
    'DiscreteSeries',
    'ModuloSeries',
    'Series',
    'SCALAR_LINEAR_INTERPOLATOR',
    'LinearInterpolationSeries',
]



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import six

from .base import DiscreteSeries, Series


def SCALAR_LINEAR_INTERPOLATOR(t0, v0, t1, v1, tt):
    """
    Good intepolator if our values can be added, subtracted, multiplied and divided
    """
    return v0 + (tt - t0) * (t1 - t0) / (v1 - v0)


class LinearInterpolationSeries(DiscreteSeries):
    def __init__(self, data, domain=None,
                 interpolator=SCALAR_LINEAR_INTERPOLATOR,
                 *args, **kwargs):
        """
        :param interpolator: callable(t0: float, v0: any, t1: float, v1: any, tt: float) -> any
            This, given intepolation points (t0, v0) and (t1, v1) such that t0 <= tt <= t1,
            return a value for index tt
        :raise TypeError: a non-discrete series was passed as data
        """
        self.interpolator = interpolator
        if isinstance(data, DiscreteSeries):
            data, domain = data.data, data.domain
        elif isinstance(data, Series):
            raise TypeError('non-discrete series not supported!')

        super(LinearInterpolationSeries, self).__init__(data, domain, *args,
                                                        **kwargs)

    def _get_for(self, item):
        if item == self.domain.start:
            return self.data[0][1]

        if len(self.data) == 1:
            return super(LinearInterpolationSeries, self).__getitem__(item)

        for i in six.moves.range(0, len(self.data) - 1):
            cur_i, cur_v = self.data[i]
            next_i, next_v = self.data[i + 1]

            if cur_i <= item <= next_i:
                return self.interpolator(cur_i, cur_v, next_i, next_v, item)

        return self.data[-1][1]



# coding=UTF-8
from __future__ import print_function, absolute_import, division

import six

from firanka.exceptions import NotInDomainError
from firanka.ranges import Range, EMPTY_SET


class Series(object):
    """
    Abstract, base class for series.

    Your series needs to override just _get_for(x: float) -> v
    for minimum functionality
    """

    def __init__(self, domain, comment=u''):
        if not isinstance(domain, Range):
            domain = Range(domain)
        self.domain = domain
        self.comment = comment

    def __getitem__(self, item):
        """
        Return a value for given index, or a subslice of this series
        :param item: a float, or a slice, or a Range
        :return: Series instance or a value
        :raises NotInDomainError: index not in domain
        """
        if isinstance(item, (Range, slice)):
            if isinstance(item, slice):
                item = Range(item)

            if item not in self.domain:
                raise NotInDomainError('slicing beyond series domain')

            newdom = self.domain.intersection(item)
            return SlicedSeries(self, newdom)
        else:
            if item not in self.domain:
                raise NotInDomainError('item not in domain')

            return self._get_for(item)

    def _get_for(self, item):
        raise NotImplementedError('This is abstract, override me!')

    def eval_points(self, points):
        """
        Return values for given points. A mass [] one could say
        :param points: iterable of indices
        :return: a list of values
        """
        return [self[p] for p in points]

    def apply(self, fun):
        """
        Return this series with a function applied to each value
        :param fun: callable/1 => 1
        :return: Series instance
        """
        return AppliedSeries(self, lambda k, v: fun(v))

    def apply_with_indices(self, fun):
        """
        Return this series with a function applied to each value
        :param fun: callable(index: float, value: any) => 1
        :return: Series instance
        """
        return AppliedSeries(self, fun)

    def discretize(self, points, domain=None):
        """
        Return this as a DiscreteSeries, sampled at points
        :return: a DiscreteSeries instance
        """
        if len(points) == 0:
            return DiscreteSeries([])

        points = list(sorted(points))

        domain = domain or Range(points[0], points[-1], True, True)

        if domain not in self.domain:
            raise NotInDomainError('points not inside this series!')

        return DiscreteSeries([(i, self[i]) for i in points], domain)

    def join(self, series, fun):
        """
        Return a new series with values of fun(v1, v2)

        :param series: series to join against
        :param fun: callable/2 => value
        :return: new Series instance
        """
        return JoinedSeries(self, series, fun)

    def translate(self, x):
        """
        Translate the series by some distance
        :param x: a float
        :return: new Series instance
        """
        return TranslatedSeries(self, x)


class AppliedSeries(Series):
    def __init__(self, series, applyfun, *args, **kwargs):
        """:type applyfun: callable(float, v) -> any"""
        super(AppliedSeries, self).__init__(series.domain, *args, **kwargs)
        self.fun = applyfun
        self.series = series

    def _get_for(self, item):
        return self.fun(item, self.series._get_for(item))


class TranslatedSeries(Series):
    def __init__(self, series, x, *args, **kwargs):
        super(TranslatedSeries, self).__init__(self.domain.translate(x), *args,
                                               **kwargs)
        self.series = series
        self.x = x

    def _get_for(self, item):
        return self.series._get_for(item + self.x)


class SlicedSeries(Series):
    def __init__(self, parent, domain, *args, **kwargs):
        super(SlicedSeries, self).__init__(domain, *args, **kwargs)
        self.parent = parent

    def _get_for(self, item):
        return self.parent._get_for(item)


def _appendif(lst, ptr, v):
    if len(lst) > 0:
        assert lst[-1][0] <= ptr
        if lst[-1][0] == ptr:
            return  # same ptr as before? Not required.
        if lst[-1][1] == v:
            return  # same value as before? Redundant
    lst.append((ptr, v))


class DiscreteSeries(Series):
    def __init__(self, data, domain=None, *args, **kwargs):
        if len(data) == 0:
            domain = EMPTY_SET
        elif domain is None:
            domain = Range(data[0][0], data[-1][0], True, True)

        self.data = data
        super(DiscreteSeries, self).__init__(domain, *args, **kwargs)

        if len(data) > 0:
            if self.domain.start < data[0][0]:
                raise ValueError(
                    'some domain space is not covered by definition!')

    def apply(self, fun):
        return DiscreteSeries([(k, fun(v)) for k, v in self.data], self.domain)

    def apply_with_indices(self, fun):
        return DiscreteSeries([(k, fun(k, v)) for k, v in self.data],
                              self.domain)

    def _get_for(self, item):
        for k, v in reversed(self.data):
            if k <= item:
                return v

        raise RuntimeError('should never happen')

    def translate(self, x):
        return DiscreteSeries([(k + x, v) for k, v in self.data],
                              self.domain.translate(x))

    def _join_discrete_other_discrete(self, series, fun):
        new_domain = self.domain.intersection(series.domain)

        assert isinstance(series, DiscreteSeries)

        a = self.data[::-1]
        b = series.data[::-1]

        ptr = self.domain.start
        c = [(ptr, fun(self._get_for(ptr), series._get_for(ptr)))]

        while len(a) > 0 and len(b) > 0:
            if a[-1] < b[-1]:
                ptr, v1 = a.pop()
                v2 = series._get_for(ptr)
            elif a[-1] > b[-1]:
                ptr, v1 = b.pop()
                v2 = self._get_for(ptr)
            else:
                ptr, v1 = a.pop()
                _, v2 = b.pop()

            _appendif(c, ptr, fun(v1, v2))

        if len(a) > 0 or len(b) > 0:
            if len(a) > 0:
                rest = a
                static_v = series._get_for(ptr)
                op = lambda me, const: fun(me, const)
            else:
                rest = b
                static_v = self._get_for(ptr)
                op = lambda me, const: fun(const, me)

            for ptr, v in rest:
                _appendif(c, ptr, op(v, static_v))

        return DiscreteSeries(c, new_domain)

    def join_discrete(self, series, fun):
        new_domain = self.domain.intersection(series.domain)

        if isinstance(series, DiscreteSeries):
            return self._join_discrete_other_discrete(series, fun)

        if new_domain.start > self.data[0][0]:
            c = [(new_domain.start, fun(self._get_for(new_domain.start),
                                        series._get_for(new_domain.start)))]
        else:
            c = []

        for k, v in ((k, v) for k, v in self.data if
                     new_domain.start <= k <= new_domain.stop):
            _appendif(c, k, fun(v, series._get_for(k)))

        if c[-1][0] != new_domain.stop:
            c.append((new_domain.stop, fun(self._get_for(new_domain.stop),
                                           series._get_for(new_domain.stop))))

        return DiscreteSeries(c, new_domain)

    def compute(self):
        """Simplify self"""
        nd = [self.data[0]]
        for i in six.moves.range(1, len(self.data)):
            if self.data[i][1] != nd[-1][1]:
                nd.append(self.data[i])
        return DiscreteSeries(nd, self.domain)


class FunctionSeries(Series):
    """
    Series with values defined by a function
    """

    def __init__(self, fun, domain, *args, **kwargs):
        super(FunctionSeries, self).__init__(domain, *args, **kwargs)
        self.fun = fun

    def _get_for(self, item):
        return self.fun(item)


class JoinedSeries(Series):
    """
    Series stemming from performing an operation on two series
    """

    def __init__(self, ser1, ser2, op, *args, **kwargs):
        domain = ser1.domain.intersection(ser2.domain)
        super(JoinedSeries, self).__init__(domain, *args, **kwargs)
        self.ser1 = ser1
        self.ser2 = ser2
        self.op = op

    def _get_for(self, item):
        return self.op(self.ser1._get_for(item), self.ser2._get_for(item))
